;; Generated package description from scroll-restore.el  -*- no-byte-compile: t -*-
(define-package "scroll-restore" "1.0.0.20221221.81959" "restore original position after scrolling" 'nil :commit "b0c162e88c56afadaa9b11a581c0b74261bc194f" :url "https://elpa.gnu.org/packages/scroll-restore.html" :authors '(("Martin Rudalics" . "rudalics@gmx.at")) :maintainer '("Martin Rudalics" . "rudalics@gmx.at") :keywords '("scrolling"))
